package tools.descartes.dql.connector.simQPN;

import java.util.LinkedList;
import java.util.List;

import org.apache.log4j.Logger;
import org.dom4j.Element;

import tools.descartes.dql.core.engine.util.DQLLogger;
import de.tud.cs.qpe.model.DocumentManager;
import de.tud.cs.qpe.model.NetHelper;
import de.tud.cs.qpe.model.PlaceHelper;
import de.tud.cs.qpe.model.ProbeHelper;
import de.tud.cs.simqpn.ui.model.BatchMeansConfiguration;
import de.tud.cs.simqpn.ui.model.Configuration;

public class ConfigurationBuilder {

	private final Logger log = DQLLogger.getLogger(this.getClass().getName());
	private String configName;
	private Double warmupPeriod;
	private Double totalRunLength;
	private String stoppingRule;
	private Double timeBetweenStopChecks;
	private Double secondsBetweenStopChecks;
	private Double accuracy;
	private int verbosity;
	private String outputDirectory;
	private List<StatLevel> statLevels = new LinkedList<StatLevel>();
	
	public ConfigurationBuilder(String configName) {
		this.configName = configName;
		this.stoppingRule = Configuration.RELATIVE_PRECISION_STOPPING_RULE;
		this.timeBetweenStopChecks = 10.0;
		this.verbosity = 1;
		this.outputDirectory = ".";
		this.totalRunLength = Double.MAX_VALUE;
		this.warmupPeriod = 50000000.0;
		this.secondsBetweenStopChecks = 3.0;
	}
	
	public void addStatLevel(String placeId, int level) {
		for (StatLevel sl: statLevels)
			if (placeId.equals(sl.getPlaceId())) {
				if (level > sl.getLevel()) {
					sl.setLevel(level);
				}
				return;
			}
		statLevels.add(new StatLevel(placeId, level));
	}
	
	public Configuration buildConfiguration(Element net) {
		DocumentManager.registerInput(net.getDocument());
		Configuration newConfiguration = createConfiguration(net, configName);
		//Warmup Period
		newConfiguration.setRampUpLength(warmupPeriod);
		//Total Run time
		newConfiguration.setTotalRunLength(totalRunLength);
		//Stopping Rule
		newConfiguration.setStoppingRule(stoppingRule);
		//Time between Stop Checks
		newConfiguration.setTimeBetweenStopChecks(timeBetweenStopChecks);
		//Seconds between Stop Checks
		newConfiguration.setSecondsBetweenStopChecks(secondsBetweenStopChecks);
		//Verbosity
		newConfiguration.setVerbosityLevel(verbosity);
		//Output directory
		newConfiguration.setOutputDirectory(outputDirectory);
		
		//Set levels for selected places
		for (StatLevel sl: statLevels) {
			Element place = NetHelper.getPlaceById(net, sl.getPlaceId());
			if (place == null)
				place = NetUtil.getProbeById(net, sl.getPlaceId());
			Element	placeMeta = PlaceHelper.createSimqpnPlaceConfigurationMetadata(
					place, newConfiguration.getName());
			placeMeta.addAttribute("statsLevel", "" + sl.getLevel());
		}
			
		//Set level to 0 for the rest of the places
		for (Element place: NetHelper.listAllPlaces(net)) {
			Element placeMeta = NetHelper.getMetadata(place, configName);
			if (placeMeta == null) {
				placeMeta = PlaceHelper.createSimqpnPlaceConfigurationMetadata(place, configName);
			}
			if (placeMeta.attribute("statsLevel") == null) {
				placeMeta.addAttribute("statsLevel", "0");
			}
		}
		
		//Set level to 0 for all probes
		for (Element probe : NetHelper.listProbes(net)) {
			Element probeMeta = NetHelper.getMetadata(probe, configName);
			probeMeta = ProbeHelper.createSimqpnProbeConfigurationMetadata(probe, configName);
			probeMeta.addAttribute("statsLevel", "0");
		}
		
		for (Element place : NetHelper.listAllPlaces(net)) {
			Element placeMeta = NetHelper.getMetadata(place, configName);
			if (placeMeta == null) {
				placeMeta = PlaceHelper.createSimqpnPlaceConfigurationMetadata(place, configName);
			}
			if (placeMeta.attribute("statsLevel") == null) {
				placeMeta.addAttribute("statsLevel", "1");
			}
			
			List<Element> colorRefs = PlaceHelper.listColorReferences(place);
			for (Element colorRef : colorRefs) {
				Element colorRefMeta = NetHelper.getMetadata(colorRef, configName);
				if (colorRefMeta == null) {
					colorRefMeta = newConfiguration.createColorRefMetadata(colorRef, configName);	
				}	
				colorRefMeta.addAttribute("reqRelPrc", "" + (1.0 - accuracy));				
				newConfiguration.initColorRefDepositoryMetadata(colorRefMeta);
				if ("queueing-place".equals(place.attributeValue("type"))) {
					newConfiguration.initColorRefQueueMetadata(colorRefMeta);	
					colorRefMeta.addAttribute("queueReqRelPrc", "" + (1.0 - accuracy));				
				}
			}
		}
		return newConfiguration;
	}
	
	private Configuration createConfiguration(Element net, String name) {
		Configuration conf = new BatchMeansConfiguration();
		conf.createSimulationConfiguration(net, name);
		return conf;
	}
	
	public List<StatLevel> getStatLevels() {
		return statLevels;
	}
	
	public Double getAccuracy() {
		return accuracy;
	}
	
	public Double getWarmupPeriod() {
		return warmupPeriod;
	}
	public Double getTotalRunLength() {
		return totalRunLength;
	}

	public void setAccuracy(Double accuracy) {
		this.accuracy = accuracy;
	}

	public Double getSecondsBetweenStopChecks() {
		return secondsBetweenStopChecks;
	}

	public class StatLevel {
		private String placeId;
		private int level;
		
		public StatLevel(String placeId, int level) {
			if (level < 0 || level > 4)
				log.error("The statlevel has to be between 0 and 4!");
			this.placeId = placeId;
			this.level = level;
		}
		
		public String getPlaceId() {
			return placeId;
		}
		
		public int getLevel() {
			return level;
		}
		
		public void setLevel(int level ) {
			if (level < 0 || level > 4)
				log.error("The statlevel has to be between 0 and 4!");
			this.level = level;
		}
	}
}
