package tools.descartes.dql.connector.simQPN;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import org.apache.log4j.Logger;
import org.dom4j.Element;
import org.eclipse.emf.ecore.util.EcoreUtil;

import de.tud.cs.qpe.model.NetHelper;
import de.tud.cs.qpe.model.PlaceHelper;
import tools.descartes.dql.connector.ModelStructureQueryConnector;
import tools.descartes.dql.core.engine.util.DQLLogger;
import tools.descartes.dql.models.mapping.domain.DoF;
import tools.descartes.dql.models.mapping.domain.Probe;
import tools.descartes.dql.models.mapping.domain.Resource;
import tools.descartes.dql.models.mapping.domain.Service;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;

public class ModelStructureQueryConnectorImpl implements ModelStructureQueryConnector {

	private final Logger log = DQLLogger.getLogger(this.getClass().getName());
	
	public String getBundleProperty(String key) {
		return Activator.getContext().getProperty(key);
	}
	
	@Override
	public EntityMapping requestDoF(EntityMapping request) {
		Element net = NetUtil.loadXmlFromFile(request.getModelLocation());
		final EntityMapping response = EcoreUtil.copy(request);
		
		List<String> dofNames = getAvailibleDoFOptions(net);
		List<DoF> dofs = EntityMappingUtil.createDoFs(dofNames);
		response.getDoF().addAll(dofs);
		return response;
	}
	
	@Override
	public EntityMapping requestEntities(EntityMapping request) {
		final EntityMapping response = EcoreUtil.copy(request);
		
		//Load 
		Element net = NetUtil.loadXmlFromFile(request.getModelLocation());
		
		//Add resources
		List<String> resourceNames = getPlaceNames(net);
		resourceNames.addAll(getQueueNames(net));
		response.getServices().addAll(EntityMappingUtil.createServices(getProbeNames(net)));
		response.getResources().addAll(EntityMappingUtil.createResources(resourceNames));
		
		return response;
	}

	private List<String> getQueueNames(Element net) {
		List<String> queueNames = new LinkedList<String>();
		for (Element e: NetHelper.listQueues(net))
			queueNames.add("Queue_" + NetUtil.getNameOfElement(e));
		return queueNames;
	}

	@Override
	public EntityMapping requestProbes(EntityMapping request) {
		final EntityMapping response = EcoreUtil.copy(request);
		
		//Load 
		Element net = NetUtil.loadXmlFromFile(request.getModelLocation());
		for (Resource r: response.getResources()) {
			r.getProbes().clear();
			if (!r.getIdentifier().contains("Queue_"))
				r.getProbes().addAll(getAvailibleProbes(NetUtil.getPlaceByName(r.getIdentifier(),
						net), net));
			else 
				r.getProbes().addAll(getAvailibleProbesForQueue(NetUtil.removeQueueTag(r.getIdentifier()),
						net));
		}
		
		for (Service s: response.getServices()) {
			s.getProbes().clear();
			s.getProbes().addAll(getAvailibleProbesForProbe(NetUtil.getProbeByName(s.getIdentifier(), net), net));
		}
			
		return response;
	}

	private Collection<? extends Probe> getAvailibleProbesForQueue(
			String queueName, Element net) {
		List<PerformanceMetric> pms = PerformanceMetricsHelper.getPerformanceMetricsForQueue();

		List<Probe> probes = new LinkedList<Probe>();
		for (PerformanceMetric pm: pms) {
			probes.add(EntityMappingUtil.createProbe(pm.getDQLName()));
		}
		
		return probes;
	}

	@Override
	public boolean supportsDoFRequests() {
		return true;
	}

	@Override
	public boolean supportsEntityRequests() {
		return true;
	}

	@Override
	public boolean supportsProbeRequests() {
		return true;
	}
	
	private Collection<? extends Probe> getAvailibleProbes(Element element, Element net) {
		Collection<Probe> c = new LinkedList<Probe>();
		switch (NetUtil.getTypeOfPlace(element)) {
		case "queueing-place":
			c.addAll(getAvailibleProbesForQueue(element, net));
			break;
		case "ordinary-place":
			c.addAll(getAvailibleProbesForOrdinaryPlace(element, net));
			break;
		case "subnet-place":
			log.error("Metrics for Subnets are not yet implemented!");
			throw new DQLException("The SimQPN Connector currently does not support subnets");
		default:
			break;
		}
		return c;
	}

	private static Collection<? extends Probe> getAvailibleProbesForQueue(
			Element element, Element net) {
		List<String> colorNames = getColorNames(element, net);
		String queueName = NetUtil.getNameOfUnderlyingQueue(
				NetUtil.getNameOfElement(element), net);
		List<PerformanceMetric> metrics = PerformanceMetricsHelper.getPerformanceMetricsForQueueingPlace(
				colorNames, queueName);
		
		List<Probe> probes = new LinkedList<Probe>();
		for (PerformanceMetric pm: metrics) {
			probes.add(EntityMappingUtil.createProbe(pm.getDQLName()));
		}
		
		return probes;
	}

	private static String qualifyByRemoval(String colorName) {
		String ret = "";
		for (int i = 0; i < colorName.length(); i++) {
			String digit = colorName.substring(i, i + 1);
			if (digit.matches("\\w")) {
				ret += digit;
			}
		}
		return ret;
	}
	
	private static List<String> getColorNames(Element place, Element net) {
		List<String> colorNames = new LinkedList<String>();
		for (Element colorRef: PlaceHelper.listColorReferences(place)) {
			colorNames.add(qualifyByRemoval(NetUtil.getNameOfElement(NetHelper.getColorById(net,
					NetUtil.getReferencedColorId(colorRef)))));
		}
		return colorNames;
	}
	
	private static Collection<Probe> getAvailibleProbesForProbe(Element element, Element net) {
		List<String> colorNames = getColorNames(element, net);
		List<PerformanceMetric> metrics = PerformanceMetricsHelper.getPerformanceMetricsForProbe(colorNames);

		List<Probe> probes = new LinkedList<Probe>();
		for (PerformanceMetric pm: metrics) {
			probes.add(EntityMappingUtil.createProbe(pm.getDQLName()));
		}
		
		return probes;
	}
	
	private static Collection<Probe> getAvailibleProbesForOrdinaryPlace(Element element, Element net) {
		List<String> colorNames = getColorNames(element, net);
		List<PerformanceMetric> metrics = PerformanceMetricsHelper.getPerformanceMetricsForPlace(colorNames);
		
		List<Probe> probes = new LinkedList<Probe>();
		for (PerformanceMetric pm: metrics) {
			probes.add(EntityMappingUtil.createProbe(pm.getDQLName()));
		}
		
		return probes;
	}
	
	private static List<String> getAvailibleDoFOptions(Element net) {
		List<String> dofNames = new LinkedList<String>();
		for (Element place : NetHelper.listAllPlaces(net)) {
			for (Element colorRef: PlaceHelper.listColorReferences(place)) {
				dofNames.add(createDofIdentifier(
						NetUtil.getNameOfElement(place), 
						NetUtil.getNameOfElement(NetHelper.getColorById(net, 
								NetUtil.getReferencedColorId(colorRef)))));
			}
				
		}
		return dofNames;
	}
	
	private static String createDofIdentifier(String placeName, String colorName) {
		return placeName + "_" + colorName;
	}
	
	private static List<String> getProbeNames(Element net) {
		List<String> probeNames = new LinkedList<String>();
		for (Element e: NetHelper.listProbes(net))
			probeNames.add(NetUtil.getNameOfElement(e));
		return probeNames;
	}
	
	private static List<String> getPlaceNames(Element net) {
		List<String> placeNames = new LinkedList<String>();
		for (Element e: NetHelper.listPlaces(net))
			placeNames.add(NetUtil.getNameOfElement(e));
		return placeNames;
	}
}
