package tools.descartes.dql.connector.simQPN;

import static org.junit.Assert.*;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import org.dom4j.Element;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.junit.Test;
import org.junit.experimental.runners.Enclosed;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;

import tools.descartes.dql.connector.simQPN.testUtils.TestUtil;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;
import tools.descartes.dql.models.mapping.mapping.MappingFactory;
import tools.descartes.dql.models.mapping.mapping.impl.MappingFactoryImpl;

@RunWith(Enclosed.class)
public class PerformanceMetricsQueryConnectorImplTest {
	public static class UnparameterizedTests {

		/**
		 * See ModelStructureQueryConnectorImplTest
		 */
		@Test
		public void testAgainstExpectedResults() {
			String modelName = "ispass03-02.qpe";
			String resultName = "expectedPerformanceMetrics_ispass03-02.txt";
			String modelPath = FileUtil.getPath(modelName);
			String resultPath = FileUtil.getPath(resultName);
			
			//Check if dof requests are enabled
			PerformanceMetricsQueryConnectorImpl pmqc = new PerformanceMetricsQueryConnectorImpl();
			pmqc.setConstraint("Accurate");
			ModelStructureQueryConnectorImpl msqc = new ModelStructureQueryConnectorImpl();
			if (!msqc.supportsDoFRequests())
				fail("DoF requests are not supported!");
			
			//Build request
			MappingFactory mappingFactory = MappingFactoryImpl.init();
			EntityMapping request = mappingFactory.createEntityMapping();
			request.setModelLocation(modelPath);
			
			//Run request
			request = msqc.requestEntities(request);
			request = msqc.requestProbes(request);
			EntityMapping answer = pmqc.requestMetrics(request);

//			TestUtil.saveToFile(resultPath, TestUtil.entityMapptingToString(answer));
//			System.out.println(TestUtil.entityMapptingToString(answer));
			
			//check for expected answers
			if (!TestUtil.contentCompare(TestUtil.loadFromFile(resultPath),
					TestUtil.entityMapptingToString(answer))) {
				assertEquals(TestUtil.loadFromFile(resultPath),
						TestUtil.entityMapptingToString(answer));
			}
		}
		

		/*
		 * This test looks complicated, but is actually quite simple.
		 * It requests different metrics for the only place in the net and
		 * then checks if the statlevel were set accordingly.
		 * In a second phase it requests 2 metrics for the only place in the net
		 * and checks if the stat level was set accordingly (it should be the higher one 
		 * of the two metrics)
		 */
		@Test
		public void setAllConfigLevelsViaMetrics() {
			String modelName = "SingleQueue.qpe";
			List<PerformanceMetric> performanceMetrics = new ArrayList<PerformanceMetric>();
			List<String> names = new LinkedList<String>();
			names.add("color");
			performanceMetrics.addAll(PerformanceMetricsHelper.getPerformanceMetricsForQueueingPlace(names, "queue"));
		    
			ModelStructureQueryConnectorImpl msqc = new ModelStructureQueryConnectorImpl();
			PerformanceMetricsQueryConnectorImpl pmqc = new PerformanceMetricsQueryConnectorImpl();
			
			//get model
			String modelPath = FileUtil.getPath(modelName);
			Element net = NetUtil.loadXmlFromFile(modelPath);
			
			//Setup basic request
			MappingFactory mappingFactory = MappingFactoryImpl.init();
			EntityMapping request = mappingFactory.createEntityMapping();
			request.setModelLocation(modelPath);
			request = msqc.requestEntities(request);
			
			//With no metrics, no stat levels should be set
			ConfigurationBuilder cb0 = new ConfigurationBuilder("JUNITConfig");
			pmqc.setAllConfigLevelsViaMetrics(cb0, net, request);
			if (cb0.getStatLevels().size() != 0)
				fail("With no metrics set, all stat levels should be 0");
			
			//Iterate over all availible Metrics
			for (PerformanceMetric pm: performanceMetrics) {
				ConfigurationBuilder cb = new ConfigurationBuilder("JUNITConfig");
				EntityMapping requestWithMetric = EcoreUtil.copy(request);
				requestWithMetric.getResources().get(0).getProbes().add(
						EntityMappingUtil.createProbe(pm.getDQLName()));
				pmqc.setAllConfigLevelsViaMetrics(cb, net, requestWithMetric);
				if (cb.getStatLevels().get(0).getLevel() != pm.getStatLevel())
					fail("For the Metric " + pm.getDQLName() + " the StatLevel was: " 
				+ cb.getStatLevels().get(0).getLevel() + " expected: " + pm.getStatLevel());
			}
			
			//Metric combinations because we can
			for (PerformanceMetric metric: performanceMetrics) {
				for (PerformanceMetric metric2: performanceMetrics) {
					ConfigurationBuilder cb = new ConfigurationBuilder("JUNITConfig");
					EntityMapping requestWithMetric = EcoreUtil.copy(request);
					requestWithMetric.getResources().get(0).getProbes().add(
							EntityMappingUtil.createProbe(metric.getDQLName()));
					requestWithMetric.getResources().get(0).getProbes().add(
							EntityMappingUtil.createProbe(metric2.getDQLName()));
					pmqc.setAllConfigLevelsViaMetrics(cb, net, requestWithMetric);
					
					int expected = metric.getStatLevel();
					if (metric2.getStatLevel() > expected)
						expected = metric2.getStatLevel();
					
					if (cb.getStatLevels().get(0).getLevel() != expected)
						fail("For the Metrics " + metric.getDQLName() + " and " + metric2.getStatLevel() 
								+ " the StatLevel was: " + cb.getStatLevels().get(0).getLevel() 
								+ " expected: " + expected);
				}
			}
		}
	}

	/*
	 * This simple test checks if the Settings are set correctly.
	 * It uses a new ConfigurationBuilder, then sets the settings
	 * via the PerformanceMetricsQueryConnector (which has a state)
	 */
	@RunWith(Parameterized.class)
	public static class SetConfigSettingsViaConstraintTest {
	    @Parameters
	    public static Collection<Object[]> data() {
	        return Arrays.asList(new Object[][] {     
	        		{"Accurate", 0.95},
	        		{null, 0.95},
	        		{"Fast", 0.8}
	           });
	    }

		@Parameter
		public String name;
		
		@Parameter(value = 1)
		public Double accuracy;
		
		@Test
		public void setConfigSettingsViaConstraintTest() {
			PerformanceMetricsQueryConnectorImpl pmqc = new PerformanceMetricsQueryConnectorImpl();
			pmqc.setConstraint(name);
			ConfigurationBuilder cb = new ConfigurationBuilder("dummy");
			pmqc.setConfigSettingsViaConstraint(cb);
			assertEquals(accuracy, cb.getAccuracy());
		}
	}

}
